import abc
import json
from logging import Logger
from typing import Any, Callable, Optional, Union

import tiktoken
from pydantic import Field, validator

import marvin
import marvin.utilities.types
from marvin.utilities.messages import Message
from marvin.utilities.types import MarvinBaseModel


class StreamHandler(MarvinBaseModel, abc.ABC):
    callback: Callable[[Message], None] = None

    @abc.abstractmethod
    def handle_streaming_response(self, api_response) -> Message:
        raise NotImplementedError()


class OpenAIFunction(MarvinBaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    parameters: dict[str, Any] = {"type": "object", "properties": {}}
    fn: Optional[Callable] = Field(None, exclude=True)
    args: Optional[dict] = None
    """
    Base class for representing a function that can be called by an LLM. The
    format is identical to OpenAI's Functions API.

    Args:
        name (str): The name of the function. description (str): The description
        of the function. parameters (dict): The parameters of the function. fn
        (Callable): The function to be called. args (dict): The arguments to be
        passed to the function.
    """

    @classmethod
    def from_function(cls, fn: Callable, **kwargs):
        return cls(
            name=kwargs.get("name", fn.__name__),
            description=kwargs.get("description", fn.__doc__ or ""),
            parameters=marvin.utilities.types.function_to_schema(fn),
            fn=fn,
        )

    async def query(self, q: str, model: "ChatLLM" = None):
        if not model:
            model = chat_llm()
        self.args = json.loads(
            (
                await model.run(
                    messages=[Message(role="USER", content=q)],
                    functions=[self],
                    function_call={"name": self.name},
                )
            )
            .data.get("function_call")
            .get("arguments")
        )
        return self


class ChatLLM(MarvinBaseModel, abc.ABC):
    name: Optional[str] = None
    model: str
    max_tokens: int = Field(default_factory=lambda: marvin.settings.llm_max_tokens)
    temperature: float = Field(default_factory=lambda: marvin.settings.llm_temperature)

    @validator("name", always=True)
    def default_name(cls, v):
        if v is None:
            v = cls.__name__
        return v

    @property
    def context_size(self) -> int:
        return 4096

    def get_tokens(self, text: str, **kwargs) -> list[int]:
        try:
            enc = tiktoken.encoding_for_model(self.model)
        # fallback to the gpt-3.5-turbo tokenizer if the model is not found
        # note this will give the wrong answer for non-OpenAI models
        except KeyError:
            enc = tiktoken.encoding_for_model("gpt-3.5-turbo")
        return enc.encode(text)

    async def __call__(self, messages, *args, **kwargs):
        return await self.run(messages, *args, **kwargs)

    @abc.abstractmethod
    def format_messages(
        self, messages: list[Message]
    ) -> Union[str, dict, list[Union[str, dict]]]:
        """Format Marvin message objects into a prompt compatible with the LLM model"""
        return messages

    @abc.abstractmethod
    async def run(
        self,
        messages: list[Message],
        functions: list[OpenAIFunction] = None,
        *,
        logger: Logger = None,
        stream_handler: Callable[[Message], None] = False,
        **kwargs,
    ) -> Message:
        """Run the LLM model on a list of messages and optional list of functions"""
        raise NotImplementedError()


def chat_llm(model: str = None, **kwargs) -> ChatLLM:
    """Dispatches to all supported LLM providers"""
    if model is None:
        model = marvin.settings.llm_model

    # automatically detect gpt-3.5 and gpt-4 for backwards compatibility
    if model.startswith("gpt-3.5-turbo") or model.startswith("gpt-4"):
        model = f"openai/{model}"

    # extract the provider and model name
    provider, model_name = model.split("/", 1)

    if provider == "openai":
        from .openai import OpenAIChatLLM

        return OpenAIChatLLM(model=model_name, **kwargs)
    elif provider == "anthropic":
        from .anthropic import AnthropicChatLLM

        return AnthropicChatLLM(model=model_name, **kwargs)
    elif provider == "azure_openai":
        from .azure_openai import AzureOpenAIChatLLM

        return AzureOpenAIChatLLM(model=model_name, **kwargs)
    else:
        raise ValueError(f"Unknown provider/model: {model}")



import inspect
import json
from types import FunctionType
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    overload,
)

from marvin._compat import cast_to_json, model_dump
from marvin.utilities.async_utils import run_sync
from marvin.utilities.logging import get_logger
from marvin.utilities.messages import Message, Role
from pydantic import BaseModel, Field
from typing_extensions import ParamSpec

from .utils import parse_raw

T = TypeVar(
    "T",
    bound=BaseModel,
)

P = ParamSpec("P")


class Request(BaseModel, Generic[T], extra="allow", arbitrary_types_allowed=True):
    messages: Optional[list[Message]] = Field(default=None)
    functions: Optional[list[Union[Callable[..., Any], dict[str, Any]]]] = Field(
        default=None
    )
    function_call: Any = None
    response_model: Optional[type[T]] = Field(default=None, exclude=True)

    def serialize(
        self,
        functions_serializer: Callable[
            [Callable[..., Any]], dict[str, Any]
        ] = cast_to_json,
    ) -> dict[str, Any]:
        extras = model_dump(
            self, exclude={"functions", "function_call", "response_model"}
        )
        response_model: dict[str, Any] = {}
        functions: dict[str, Any] = {}
        function_call: dict[str, Any] = {}
        messages: dict[str, Any] = {}

        if self.response_model:
            response_model_schema: dict[str, Any] = functions_serializer(
                self.response_model
            )
            response_model = {
                "functions": [response_model_schema],
                "function_call": {"name": response_model_schema.get("name")},
            }

        elif self.functions:
            functions = {
                "functions": [
                    functions_serializer(function) for function in self.functions
                ]
            }
            if self.function_call:
                functions["function_call"] = self.function_call

        return extras | functions | function_call | messages | response_model

    def function_registry(self) -> dict[str, FunctionType]:
        return {
            function.__name__: function
            for function in self.functions or []
            if callable(function)
        }


class Choice(BaseModel):
    message: Message
    index: int
    finish_reason: str


class Usage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class Response(BaseModel, Generic[T], extra="allow", arbitrary_types_allowed=True):
    id: str
    object: str
    created: int
    model: str
    usage: Usage
    choices: list[Choice] = Field(default_factory=list)


class Turn(BaseModel, Generic[T], extra="allow", arbitrary_types_allowed=True):
    request: Request[T]
    response: Response[T]

    @overload
    def __getitem__(self, key: Literal[0]) -> Request[T]:
        ...

    @overload
    def __getitem__(self, key: Literal[1]) -> Response[T]:
        ...

    def __getitem__(self, key: int) -> Union[Request[T], Response[T]]:
        if key == 0:
            return self.request
        elif key == 1:
            return self.response
        else:
            raise IndexError("Turn only has two items.")

    def has_function_call(self) -> bool:
        return any([choice.message.function_call for choice in self.response.choices])

    def get_function_call(self) -> list[tuple[str, dict[str, Any]]]:
        if not self.has_function_call():
            raise ValueError("No function call found.")
        pairs: list[tuple[str, dict[str, Any]]] = []
        for choice in self.response.choices:
            if choice.message.function_call:
                pairs.append(
                    (
                        choice.message.function_call.name,
                        parse_raw(choice.message.function_call.arguments),
                    )
                )
        return pairs

    def call_function(self) -> Union[Message, list[Message]]:
        if not self.has_function_call():
            raise ValueError("No function call found.")

        pairs: list[tuple[str, dict[str, Any]]] = self.get_function_call()
        function_registry: dict[str, FunctionType] = self.request.function_registry()
        evaluations: list[Any] = []

        logger = get_logger("ChatCompletion.handlers")

        for pair in pairs:
            name, argument = pair
            if name not in function_registry:
                raise ValueError(
                    f"Function {name} not found in {function_registry=!r}."
                )

            logger.debug_kv(
                "Function call",
                (
                    f"Calling function {name!r} with payload:"
                    f" {json.dumps(argument, indent=2)}"
                ),
                key_style="green",
            )

            function_result = function_registry[name](**argument)

            if inspect.isawaitable(function_result):
                function_result = run_sync(function_result)

            logger.debug_kv(
                "Function call",
                f"Function {name!r} returned: {function_result}",
                key_style="green",
            )

            evaluations.append(function_result)
        if len(evaluations) != 1:
            return [
                Message(
                    name=pairs[j][0],
                    role=Role.FUNCTION_RESPONSE,
                    content=str(evaluations[j]),
                    function_call=None,
                )
                for j in range(len(evaluations))
            ]
        else:
            return Message(
                name=pairs[0][0],
                role=Role.FUNCTION_RESPONSE,
                content=str(evaluations[0]),
                function_call=None,
            )

    def to_model(self) -> T:
        if not self.request.response_model:
            raise ValueError("No response model found.")
        model = self.request.response_model
        pairs = self.get_function_call()
        try:
            return model(**pairs[0][1])
        except TypeError:
            pass
        try:
            return model.parse_raw(pairs[0][1])  # type: ignore
        except TypeError:
            pass
        return model.construct(**pairs[0][1])  # type: ignore



import inspect
import json
import re
from logging import Logger
from typing import Callable, Union

import anthropic
import openai
import openai.openai_object
from pydantic import BaseModel

import marvin
import marvin.utilities.types
from marvin.engine.language_models import ChatLLM, StreamHandler
from marvin.engine.language_models.base import OpenAIFunction
from marvin.utilities.async_utils import create_task
from marvin.utilities.logging import get_logger
from marvin.utilities.messages import Message, Role
from marvin.utilities.strings import jinja_env

CONTEXT_SIZES = {
    "claude-instant": 100_000,
    "claude-2": 100_000,
}

FUNCTION_CALL_REGEX = re.compile(
    r'{\s*"mode":\s*"function_call"\s*(.*)}',
    re.DOTALL,
)
FUNCTION_CALL_NAME = re.compile(r'"name":\s*"(.*)"')
FUNCTION_CALL_ARGS = re.compile(r'"arguments":\s*(".*")', re.DOTALL)


def extract_function_call(completion: str) -> Union[dict, None]:
    function_call = dict(name=None, arguments="{}")
    if match := FUNCTION_CALL_REGEX.search(completion):
        if name := FUNCTION_CALL_NAME.search(match.group(1)):
            function_call["name"] = name.group(1)
        if args := FUNCTION_CALL_ARGS.search(match.group(1)):
            function_call["arguments"] = args.group(1)
            try:
                function_call["arguments"] = json.loads(function_call["arguments"])
            except json.JSONDecodeError:
                pass
    if not function_call["name"]:
        return None
    return function_call


def anthropic_role_map(marvin_role: Role):
    if marvin_role in [Role.USER, Role.SYSTEM, Role.FUNCTION_RESPONSE]:
        return anthropic.HUMAN_PROMPT
    else:
        return anthropic.AI_PROMPT


class AnthropicFunctionCall(BaseModel):
    mode: str
    name: str
    arguments: str

    @classmethod
    def parse_raw(cls, raw: str):
        return super().parse_raw(re.sub("^[^{]*|[^}]*$", "", raw))


class AnthropicStreamHandler(StreamHandler):
    async def handle_streaming_response(
        self,
        api_response: openai.openai_object.OpenAIObject,
    ) -> Message:
        """
        Accumulate chunk deltas into a full response. Returns the full message.
        Passes partial messages to the callback, if provided.
        """
        response = {
            "role": Role.ASSISTANT,
            "content": "",
            "data": {},
            "llm_response": None,
        }

        async for msg in api_response:
            response["llm_response"] = msg.dict()
            response["content"] += msg.completion

            if function_call := extract_function_call(response["content"]):
                response["role"] = Role.FUNCTION_REQUEST
                response["data"]["function_call"] = function_call

            if self.callback:
                callback_result = self.callback(Message(**response))
                if inspect.isawaitable(callback_result):
                    create_task(callback_result)

        response["content"] = response["content"].strip()
        return Message(**response)


class AnthropicChatLLM(ChatLLM):
    model: str = "claude-2"

    @property
    def context_size(self) -> int:
        if self.model in CONTEXT_SIZES:
            return CONTEXT_SIZES[self.model]
        else:
            for model_prefix, context in CONTEXT_SIZES:
                if self.model.startswith(model_prefix):
                    return context
        return 100_000

    def format_messages(
        self, messages: list[Message]
    ) -> Union[str, dict, list[Union[str, dict]]]:
        formatted_messages = []
        for msg in messages:
            role = anthropic_role_map(msg.role)
            formatted_messages.append(f"{role}{msg.content}")

        return "".join(formatted_messages) + anthropic.AI_PROMPT

    async def run(
        self,
        messages: list[Message],
        *,
        functions: list[OpenAIFunction] = None,
        function_call: Union[str, dict[str, str]] = None,
        logger: Logger = None,
        stream_handler: Callable[[Message], None] = False,
        **kwargs,
    ) -> Message:
        """Calls an OpenAI LLM with a list of messages and returns the response."""

        if logger is None:
            logger = get_logger(self.name)

        # ----------------------------------
        # Prepare functions
        # ----------------------------------
        if functions:
            function_message = jinja_env.from_string(FUNCTIONS_INSTRUCTIONS).render(
                functions=functions, function_call=function_call
            )
            system_message = Message(role=Role.SYSTEM, content=function_message)
            messages = [system_message] + messages

        prompt = self.format_messages(messages)

        # ----------------------------------
        # Call Anthropic LLM
        # ----------------------------------

        if not marvin.settings.anthropic.api_key:
            raise ValueError(
                "Anthropic API key not found in settings. Please set it or use the"
                " MARVIN_ANTHROPIC_API_KEY environment variable."
            )

        client = anthropic.AsyncAnthropic(
            api_key=marvin.settings.anthropic.api_key.get_secret_value(),
            timeout=marvin.settings.llm_request_timeout_seconds,
        )

        kwargs.setdefault("temperature", self.temperature)
        kwargs.setdefault("max_tokens_to_sample", self.max_tokens)

        response = await client.completions.create(
            model=self.model,
            prompt=prompt,
            stream=True if stream_handler else False,
            **kwargs,
        )

        if stream_handler:
            handler = AnthropicStreamHandler(callback=stream_handler)
            msg = await handler.handle_streaming_response(response)
            return msg

        else:
            llm_response = response.dict()
            content = llm_response["completion"].strip()
            role = Role.ASSISTANT
            data = {}
            if function_call := extract_function_call(content):
                role = Role.FUNCTION_REQUEST
                data["function_call"] = function_call
            msg = Message(
                role=role,
                content=content,
                data=data,
                llm_response=llm_response,
            )
            return msg


FUNCTIONS_INSTRUCTIONS = """
# Functions

You can call various functions to perform tasks.

Whenever you receive a message from the user, check to see if any of your
functions would help you respond. For example, you might use a function to look
up information, interact with a filesystem, call an API, or validate data. You
might write code, update state, or cause a side effect. After indicating that
you want to call a function, the user will execute the function and tell you its
result so that you can use the information in your final response. Therefore,
you must use your functions whenever they would be helpful.

The user may also provide a `function_call` instruction which could be:

- "auto": you may decide to call a function on your own (this is the
    default)
- "none": do not call any function
- {"name": "<function-name>"}: you MUST call the function with the given
    name

To call a function:

- Your response must include a JSON payload with the below format, including the
  {"mode": "function_call"} key.
- Do not put any other text in your response beside the JSON payload.
- Do not describe your plan to call the function to the user; they will not see
  it. 
- Do not include more than one payload in your response.
- Do not include function output or results in your response.
    
# Available Functions

Your have access to the following functions. Each has a name (which must be part
of your response), a description (which you should use to decide to call the
function), and a parameter spec (which is a JSON Schema description of the
arguments you should pass in your response)

{% for function in functions -%} 

## {{ function.name }} 

- Name: {{ function.name }} 
- Description: {{ function.description }} 
- Parameters: {{ function.parameters }}

{% endfor %}

# Calling a Function

To call a function, your response MUST include a JSON document with the
following structure: 

{
    "mode": "function_call", 
    
    "name": "<the name of the function, must be one of the above names>", 
    
    "arguments": "<a JSON string with the arguments to the function as valid
    JSON>"
}

The user will execute the function and respond with its result verbatim.

# function_call instruction

The user provided the following `function_call` instruction: {{ function_call }}
"""



from pathlib import Path

import marvin
import marvin.tools
import marvin.tools.filesystem
import marvin.tools.python
import marvin.tools.shell
from marvin import AIApplication
from pydantic import BaseModel, Field

marvin.settings.log_level = "DEBUG"
marvin.settings.llm_model = "gpt-4"


class TestWriterState(BaseModel):
    files_info: dict[str, str] = Field(
        default_factory=dict,
        description=(
            "A place to record notes about specific files, including any details e.g."
            ' {"path/to/file.py": "Main entrypoint"}'
        ),
    )
    tests_passing: bool = False


ROOT_DIR = Path(marvin.__file__).parents[2]

test_app = AIApplication(
    name="TestWriter",
    description=f"""
        You are responsible for writing and maintaining the unit tests for the
        Marvin library, located at {ROOT_DIR}. You may only modify files inside
        {ROOT_DIR}. Marvin's tests are all stored in `tests/` and can be run
        with `pytest` from the directory root. 
        
        The user will give you instructions on what functionality to test or how
        to modify tests. When you write tests, you will need to ensure that the
        work and meet the user's expectation. Remember, you are an expert Python
        developer and you strive to write complete, readable, tests. You believe
        tests are the best form of documentation. Do not write tests that
        already exist, make sure you are adding valuable and interesting tests
        to the codebase.
        """,
    state=TestWriterState(),
    tools=[
        marvin.tools.filesystem.ListFiles(root_dir=ROOT_DIR),
        marvin.tools.filesystem.ReadFile(root_dir=ROOT_DIR),
        marvin.tools.filesystem.ReadFiles(root_dir=ROOT_DIR),
        marvin.tools.filesystem.WriteFiles(
            root_dir=ROOT_DIR, require_confirmation=False
        ),
        marvin.tools.python.Python(require_confirmation=False),
        marvin.tools.shell.Shell(
            require_confirmation=False, working_directory=ROOT_DIR
        ),
    ],
)



from rich.prompt import Prompt
from rich.console import Console
from rich.panel import Panel
from rich import box


def _reset_history():
    global history
    history = []
    console.print(Panel("History has been reset.", box=box.DOUBLE_EDGE, expand=False))


def _get_settings():
    import marvin.settings

    console.print(
        Panel(
            f"Settings:\n{marvin.settings.json(indent=2)}",
            box=box.DOUBLE_EDGE,
            expand=False,
        )
    )


KNOWN_COMMANDS = {
    "!refresh": _reset_history,
    "!settings": _get_settings,
}

console = Console()


def format_user_input(user_input):
    return f"[bold blue]You:[/bold blue] {user_input}"


def format_chatbot_response(response):
    return f"[bold green]Marvin:[/bold green] {response}"


async def chat():
    console.print(
        Panel(
            "[bold]Welcome to the Marvin Chat CLI![/bold]", box=box.DOUBLE, expand=False
        )
    )
    console.print(
        Panel("You can type 'quit' or 'exit' to end the conversation.", expand=False)
    )
    from marvin.engine.language_models import chat_llm
    from marvin.utilities.messages import Message

    global history
    history = []
    model = chat_llm()
    try:
        while True:
            user_input = Prompt.ask("❯ ")
            if (input_lower := user_input.lower()) in ["quit", "exit"]:
                break
            if input_lower in KNOWN_COMMANDS:
                KNOWN_COMMANDS[input_lower]()
                continue

            with console.status("[bold green]Processing...", spinner="dots"):
                user_message = Message(role="USER", content=user_input)
                response = await model.run(messages=history + [user_message])
                history.extend([user_message, response])
            console.print(
                Panel(
                    format_chatbot_response(response.content),
                    box=box.ROUNDED,
                    expand=False,
                ),
            )
    except KeyboardInterrupt:
        pass



from .settings import settings

from .components import (
    ai_classifier,
    ai_fn,
    ai_model,
    AIApplication,
    AIFunction,
    AIModel,
    AIModelFactory,
)

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"


from .core.ChatCompletion import ChatCompletion

__all__ = [
    "ai_classifier",
    "ai_fn",
    "ai_model",
    "AIApplication",
    "AIFunction",
    "AIModel",
    "AIModelFactory",
    "settings",
]



import os
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Literal, Optional, Union

from ._compat import (
    BaseSettings,
    Field,
    SecretStr,
    model_dump,
)

DEFAULT_ENV_PATH = Path(os.getenv("MARVIN_ENV_FILE", "~/.marvin/.env")).expanduser()


class MarvinBaseSettings(BaseSettings):
    class Config:
        env_file = (
            ".env",
            str(DEFAULT_ENV_PATH),
        )
        env_prefix = "MARVIN_"
        validate_assignment = True


class OpenAISettings(MarvinBaseSettings):
    """Provider-specific settings. Only some of these will be relevant to users."""

    class Config:
        env_prefix = "MARVIN_OPENAI_"

    api_key: Optional[SecretStr] = None
    organization: Optional[str] = None
    embedding_engine: str = "text-embedding-ada-002"
    api_type: Optional[str] = None
    api_base: Optional[str] = None
    api_version: Optional[str] = None

    def get_defaults(self, settings: "Settings") -> dict[str, Any]:
        import os

        import openai

        from marvin import openai as marvin_openai

        EXCLUDE_KEYS = {"stream_handler"}

        response: dict[str, Any] = {}
        if settings.llm_max_context_tokens > 0:
            response["max_tokens"] = settings.llm_max_tokens
        response["api_key"] = self.api_key and self.api_key.get_secret_value()
        if os.environ.get("MARVIN_OPENAI_API_KEY"):
            response["api_key"] = os.environ["MARVIN_OPENAI_API_KEY"]
        if os.environ.get("OPENAI_API_KEY"):
            response["api_key"] = os.environ["OPENAI_API_KEY"]
        if openai.api_key:
            response["api_key"] = openai.api_key
        if marvin_openai.api_key:
            response["api_key"] = marvin_openai.api_key
        response["temperature"] = settings.llm_temperature
        response["request_timeout"] = settings.llm_request_timeout_seconds
        return {
            k: v for k, v in response.items() if v is not None and k not in EXCLUDE_KEYS
        }


class AnthropicSettings(MarvinBaseSettings):
    class Config:
        env_prefix = "MARVIN_ANTHROPIC_"

    api_key: Optional[SecretStr] = None

    def get_defaults(self, settings: "Settings") -> dict[str, Any]:
        response: dict[str, Any] = {}
        if settings.llm_max_context_tokens > 0:
            response["max_tokens_to_sample"] = settings.llm_max_tokens
        response["api_key"] = self.api_key and self.api_key.get_secret_value()
        response["temperature"] = settings.llm_temperature
        response["timeout"] = settings.llm_request_timeout_seconds
        if os.environ.get("MARVIN_ANTHROPIC_API_KEY"):
            response["api_key"] = os.environ["MARVIN_ANTHROPIC_API_KEY"]
        if os.environ.get("ANTHROPIC_API_KEY"):
            response["api_key"] = os.environ["ANTHROPIC_API_KEY"]
        return {k: v for k, v in response.items() if v is not None}


class AzureOpenAI(MarvinBaseSettings):
    class Config:
        env_prefix = "MARVIN_AZURE_OPENAI_"

    api_key: Optional[SecretStr] = None
    api_type: Literal["azure", "azure_ad"] = "azure"
    # "The endpoint of the Azure OpenAI API. This should have the form https://YOUR_RESOURCE_NAME.openai.azure.com" # noqa
    api_base: Optional[str] = None
    api_version: Optional[str] = "2023-07-01-preview"
    # `deployment_name` will correspond to the custom name you chose for your deployment when # noqa
    # you deployed a model.
    deployment_name: Optional[str] = None

    def get_defaults(self, settings: "Settings") -> dict[str, Any]:
        import os

        import openai

        from marvin import openai as marvin_openai

        response: dict[str, Any] = {}
        if settings.llm_max_context_tokens > 0:
            response["max_tokens"] = settings.llm_max_tokens
        response["temperature"] = settings.llm_temperature
        response["request_timeout"] = settings.llm_request_timeout_seconds
        response["api_key"] = self.api_key and self.api_key.get_secret_value()
        if os.environ.get("MARVIN_AZURE_OPENAI_API_KEY"):
            response["api_key"] = os.environ["MARVIN_AZURE_OPENAI_API_KEY"]
        if openai.api_key:
            response["api_key"] = openai.api_key
        if marvin_openai.api_key:
            response["api_key"] = marvin_openai.api_key

        return model_dump(self, exclude_unset=True) | {
            k: v for k, v in response.items() if v is not None
        }


def initial_setup(home: Union[Path, None] = None) -> Path:
    if not home:
        home = Path.home() / ".marvin"
    home.mkdir(parents=True, exist_ok=True)
    return home


class Settings(MarvinBaseSettings):
    """Marvin settings"""

    home: Path = initial_setup()
    test_mode: bool = False

    # LOGGING
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO"
    verbose: bool = False

    # LLMS
    llm_model: str = "openai/gpt-3.5-turbo"
    llm_max_tokens: int = 1500
    llm_max_context_tokens: int = 3500
    llm_temperature: float = 0.8
    llm_request_timeout_seconds: Union[float, list[float]] = 600.0

    # AI APPLICATIONS
    ai_application_max_iterations: Optional[int] = None

    # providers
    openai: OpenAISettings = OpenAISettings()
    anthropic: AnthropicSettings = AnthropicSettings()
    azure_openai: AzureOpenAI = AzureOpenAI()

    # SLACK
    slack_api_token: Optional[SecretStr] = Field(
        default=None,
        description="The Slack API token to use for the Slack client",
    )

    # TOOLS

    # chroma
    chroma_server_host: Optional[str] = Field(default=None)
    chroma_server_http_port: Optional[int] = Field(default=None)

    # github
    github_token: Optional[SecretStr] = Field(default=None)

    # wolfram
    wolfram_app_id: Optional[SecretStr] = Field(default=None)

    def get_defaults(self, provider: Optional[str] = None) -> dict[str, Any]:
        response: dict[str, Any] = {}
        if provider == "openai":
            return self.openai.get_defaults(self)
        elif provider == "anthropic":
            return self.anthropic.get_defaults(self)
        elif provider == "azure_openai":
            return self.azure_openai.get_defaults(self)
        else:
            return response


settings = Settings()


@contextmanager
def temporary_settings(**kwargs: Any):
    """
    Temporarily override Marvin setting values. This will _not_ mutate values that have
    been already been accessed at module load time.

    This function should only be used for testing.

    Example:
        >>> from marvin.settings import settings
        >>> with temporary_settings(MARVIN_LLM_MAX_TOKENS=100):
        >>>    assert settings.llm_max_tokens == 100
        >>> assert settings.llm_max_tokens == 1500
    """
    old_env = os.environ.copy()
    old_settings = settings.copy()

    try:
        for setting in kwargs:
            value = kwargs.get(setting)
            if value is not None:
                os.environ[setting] = str(value)
            else:
                os.environ.pop(setting, None)

        new_settings = Settings()

        for field in settings.__fields__:
            object.__setattr__(settings, field, getattr(new_settings, field))

        yield settings
    finally:
        for setting in kwargs:
            value = old_env.get(setting)
            if value is not None:
                os.environ[setting] = value
            else:
                os.environ.pop(setting, None)

        for field in settings.__fields__:
            object.__setattr__(settings, field, getattr(old_settings, field))



import inspect

import pytest
from marvin.components.ai_function import ai_fn

from tests.utils.mark import pytest_mark_class


@ai_fn
def list_fruit(n: int = 2) -> list[str]:
    """Returns a list of `n` fruit"""


@ai_fn
def list_fruit_color(n: int, color: str = None) -> list[str]:
    """Returns a list of `n` fruit that all have the provided `color`"""


@pytest_mark_class("llm")
class TestAIFunctions:
    def test_list_fruit(self):
        result = list_fruit()
        assert len(result) == 2

    def test_list_fruit_argument(self):
        result = list_fruit(5)
        assert len(result) == 5

    async def test_list_fruit_async(self):
        @ai_fn
        async def list_fruit(n: int) -> list[str]:
            """Returns a list of `n` fruit"""

        coro = list_fruit(3)
        assert inspect.iscoroutine(coro)
        result = await coro
        assert len(result) == 3

    def test_list_fruit_with_generic_type_hints(self):
        @ai_fn
        def list_fruit(n: int) -> list[str]:
            """Returns a list of `n` fruit"""

        result = list_fruit(3)
        assert len(result) == 3


@pytest_mark_class("llm")
class TestAIFunctionsMap:
    def test_map(self):
        result = list_fruit_color.map([2, 3])
        assert len(result) == 2
        assert len(result[0]) == 2
        assert len(result[1]) == 3

    async def test_amap(self):
        result = await list_fruit_color.amap([2, 3])
        assert len(result) == 2
        assert len(result[0]) == 2
        assert len(result[1]) == 3

    def test_map_kwargs(self):
        result = list_fruit_color.map(n=[2, 3])
        assert len(result) == 2
        assert len(result[0]) == 2
        assert len(result[1]) == 3

    def test_map_kwargs_and_args(self):
        result = list_fruit_color.map([2, 3], color=[None, "red"])
        assert len(result) == 2
        assert len(result[0]) == 2
        assert len(result[1]) == 3

    def test_invalid_args(self):
        with pytest.raises(TypeError):
            list_fruit_color.map(2, color=["orange", "red"])

    def test_invalid_kwargs(self):
        with pytest.raises(TypeError):
            list_fruit_color.map([2, 3], color=None)

    async def test_invalid_async_map(self):
        with pytest.raises(TypeError, match="can't be used in 'await' expression"):
            await list_fruit_color.map(n=[2], color=["orange", "red"])



import abc
from typing import Callable, Optional

from marvin.utilities.messages import Message
from marvin.utilities.types import MarvinBaseModel


class StreamHandler(MarvinBaseModel, abc.ABC):
    callback: Optional[Callable] = None

    @abc.abstractmethod
    def handle_streaming_response(self, api_response) -> Message:
        raise NotImplementedError()



import inspect
from logging import Logger
from typing import Callable, Optional, Union

import openai
import openai.openai_object

import marvin
import marvin.utilities.types
from marvin.utilities.async_utils import create_task
from marvin.utilities.logging import get_logger
from marvin.utilities.messages import Message, Role

from .base import ChatLLM, OpenAIFunction, StreamHandler

CONTEXT_SIZES = {
    "gpt-3.5-turbo-16k-0613": 16384,
    "gpt-3.5-turbo-16k": 16384,
    "gpt-3.5-turbo-0613": 4096,
    "gpt-3.5-turbo": 4096,
    "gpt-4-32k-0613": 32768,
    "gpt-4-32k": 32768,
    "gpt-4-0613": 8192,
    "gpt-4": 8192,
}


def openai_role_map(marvin_role: Role) -> str:
    if marvin_role == Role.FUNCTION_RESPONSE or marvin_role == "function_request":
        return "function"
    elif marvin_role == Role.FUNCTION_REQUEST or marvin_role == "function_response":
        return "assistant"
    else:
        return getattr(marvin_role, "value", marvin_role).lower()


class OpenAIStreamHandler(StreamHandler):
    async def handle_streaming_response(
        self,
        api_response: openai.openai_object.OpenAIObject,
    ) -> Message:
        """
        Accumulate chunk deltas into a full response. Returns the full message.
        Passes partial messages to the callback, if provided.
        """
        response = {"role": None, "content": "", "data": {}, "llm_response": None}

        async for r in api_response:
            response["llm_response"] = r.to_dict_recursive()

            delta = r.choices[0].delta if r.choices and r.choices[0] else None

            if delta is None:
                continue

            if "role" in delta:
                response["role"] = delta.role

            if fn_call := delta.get("function_call"):
                if "function_call" not in response["data"]:
                    response["data"]["function_call"] = {"name": None, "arguments": ""}
                if "name" in fn_call:
                    response["data"]["function_call"]["name"] = fn_call.name
                if "arguments" in fn_call:
                    response["data"]["function_call"]["arguments"] += (
                        fn_call.arguments or ""
                    )

            if "content" in delta:
                response["content"] += delta.content or ""

            if self.callback:
                callback_result = self.callback(Message(**response))
                if inspect.isawaitable(callback_result):
                    create_task(callback_result)

        return Message(**response)


class OpenAIChatLLM(ChatLLM):
    model: Optional[str] = "gpt-3.5-turbo"

    @property
    def context_size(self) -> int:
        if self.model in CONTEXT_SIZES:
            return CONTEXT_SIZES[self.model]
        else:
            for model_prefix, context in CONTEXT_SIZES:
                if self.model.startswith(model_prefix):
                    return context
        return 4096

    def _get_openai_settings(self) -> dict:
        openai_kwargs = {}
        if marvin.settings.openai.api_key:
            openai_kwargs["api_key"] = marvin.settings.openai.api_key.get_secret_value()
        else:
            raise ValueError(
                "OpenAI API key not set. Please set it or use the"
                " MARVIN_OPENAI_API_KEY environment variable."
            )

        if marvin.settings.openai.api_type:
            openai_kwargs["api_type"] = marvin.settings.openai.api_type
        if marvin.settings.openai.api_base:
            openai_kwargs["api_base"] = marvin.settings.openai.api_base
        if marvin.settings.openai.api_version:
            openai_kwargs["api_version"] = marvin.settings.openai.api_version
        if marvin.settings.openai.organization:
            openai_kwargs["organization"] = marvin.settings.openai.organization
        return openai_kwargs

    def format_messages(
        self, messages: list[Message]
    ) -> Union[str, dict, list[Union[str, dict]]]:
        """Format Marvin message objects into a prompt compatible with the LLM model"""
        formatted_messages = []
        for m in messages:
            role = openai_role_map(m.role)
            fmt = {"role": role, "content": m.content}
            if m.name:
                fmt["name"] = m.name
            formatted_messages.append(fmt)
        return formatted_messages

    async def run(
        self,
        messages: list[Message],
        *,
        functions: list[OpenAIFunction] = None,
        function_call: Union[str, dict[str, str]] = None,
        logger: Logger = None,
        stream_handler: Callable[[Message], None] = False,
        **kwargs,
    ) -> Message:
        """Calls an OpenAI LLM with a list of messages and returns the response."""

        # ----------------------------------
        # Validate arguments
        # ----------------------------------

        if functions is None:
            functions = []
        if function_call is None:
            function_call = "auto"
        elif function_call not in (
            ["auto", "none"] + [{"name": f.name} for f in functions]
        ):
            raise ValueError(f"Invalid function_call value: {function_call}")
        if logger is None:
            logger = get_logger(self.name)

        # ----------------------------------
        # Form OpenAI-specific arguments
        # ----------------------------------

        openai_kwargs = self._get_openai_settings()
        kwargs.update(openai_kwargs)

        prompt = self.format_messages(messages)
        llm_functions = [f.dict(exclude={"fn"}, exclude_none=True) for f in functions]

        # only add to kwargs if supplied, because empty parameters are not
        # allowed by OpenAI
        if functions:
            kwargs["functions"] = llm_functions
            kwargs["function_call"] = function_call

        # ----------------------------------
        # Call OpenAI LLM
        # ----------------------------------

        kwargs.setdefault("temperature", self.temperature)
        kwargs.setdefault("max_tokens", self.max_tokens)

        response = await openai.ChatCompletion.acreate(
            model=self.model,
            messages=prompt,
            stream=True if stream_handler else False,
            request_timeout=marvin.settings.llm_request_timeout_seconds,
            **kwargs,
        )

        if stream_handler:
            handler = OpenAIStreamHandler(callback=stream_handler)
            msg = await handler.handle_streaming_response(response)
            role = msg.role

            if role == Role.ASSISTANT and isinstance(
                msg.data.get("function_call"), dict
            ):
                role = Role.FUNCTION_REQUEST

            return Message(
                role=role,
                content=msg.content,
                data=msg.data,
                llm_response=msg.llm_response,
            )

        else:
            llm_response = response.to_dict_recursive()
            msg = llm_response["choices"][0]["message"].copy()
            role = msg.pop("role").upper()
            if role == "ASSISTANT" and isinstance(msg.get("function_call"), dict):
                role = Role.FUNCTION_REQUEST
            msg = Message(
                role=role,
                content=msg.pop("content", None),
                data=msg,
                llm_response=llm_response,
            )
            return msg



from typing import List, Literal, Optional

import pytest
from marvin import ai_model
from pydantic import BaseModel, Field

from tests.utils.mark import pytest_mark_class


@pytest_mark_class("llm")
class TestAIModels:
    def test_arithmetic(self):
        @ai_model
        class Arithmetic(BaseModel):
            sum: float = Field(
                ..., description="The resolved sum of provided arguments"
            )
            is_odd: bool

        x = Arithmetic("One plus six")
        assert x.sum == 7
        assert x.is_odd

    def test_geospatial(self):
        @ai_model
        class Location(BaseModel):
            latitude: float
            longitude: float
            city: str
            state: str
            country: str

        x = Location("The capital city of the Cornhusker State.")
        assert x.city == "Lincoln"
        assert x.state == "Nebraska"
        assert "United" in x.country
        assert "States" in x.country
        assert x.latitude // 1 == 40
        assert x.longitude // 1 == -97

    def test_depth(self):
        from typing import List

        class Country(BaseModel):
            name: str

        class City(BaseModel):
            name: str
            country: Country

        class Neighborhood(BaseModel):
            name: str
            city: City

        @ai_model
        class RentalHistory(BaseModel):
            neighborhood: List[Neighborhood]

        assert RentalHistory("""\
            I lived in Palms, then Mar Vista, then Pico Robertson.
        """)

    def test_resume(self):
        class Experience(BaseModel):
            technology: str
            years_of_experience: int
            supporting_phrase: Optional[str]

        @ai_model
        class Resume(BaseModel):
            """Details about a person's work experience."""

            greater_than_three_years_management_experience: bool
            greater_than_ten_years_management_experience: bool
            technologies: List[Experience]

        x = Resume("""\
            Data Engineering Manager, 2017-2022
            • Managed team of three engineers and data scientists
            • Deployed and maintained internal Apache Kafka pipeline
            • Built tree-based classifier to predict customer churn (xgboost)\
        """)

        assert x.greater_than_three_years_management_experience
        assert not x.greater_than_ten_years_management_experience
        assert len(x.technologies) == 2

    @pytest.mark.flaky(reruns=2)
    def test_literal(self):
        class CertainPerson(BaseModel):
            name: Literal["Adam", "Nate", "Jeremiah"]

        @ai_model
        class LLMConference(BaseModel):
            speakers: List[CertainPerson]

        x = LLMConference("""
            The conference for best LLM framework will feature talks by
            Adam, Nate, Jeremiah, Marvin, and Billy Bob Thornton.
        """)
        assert len(set([speaker.name for speaker in x.speakers])) == 3
        assert set([speaker.name for speaker in x.speakers]) == set(
            ["Adam", "Nate", "Jeremiah"]
        )

    @pytest.mark.xfail(reason="regression in OpenAI function-using models")
    def test_history(self):
        from typing import List

        class Location(BaseModel):
            city: str
            state: str

        class Candidate(BaseModel):
            name: str
            political_party: str
            campaign_slogan: str
            birthplace: Location

        @ai_model
        class Election(BaseModel):
            candidates: List[Candidate]
            winner: Candidate

        x = Election("The United States Election of 1800")

        assert x.winner in x.candidates
        assert x.winner.name == "Thomas Jefferson"
        assert x.winner.political_party == "Democratic-Republican"
        assert x.winner.birthplace.city == "Shadwell"
        assert x.winner.birthplace.state == "Virginia"
        assert set([candidate.name for candidate in x.candidates]).issubset(
            set(
                [
                    "Thomas Jefferson",
                    "John Adams",
                    "Aaron Burr",
                    "Charles C. Pinckney",
                    "Charles Pinckney",
                    "Charles Cotesworth Pinckney",
                ]
            )
        )


@pytest_mark_class("llm")
class TestAIModelsMessage:
    def test_arithmetic_message(self):
        @ai_model
        class Arithmetic(BaseModel):
            sum: float = Field(
                ..., description="The resolved sum of provided arguments"
            )

        x = Arithmetic("One plus six")
        assert x.sum == 7
        # assert isinstance(x._message, Message)
        # assert x._message.role == Role.FUNCTION_RESPONSE


@pytest_mark_class("llm")
class TestInstructions:
    # def test_instructions_error(self):
    #     @ai_model
    #     class Test(BaseModel):
    #         text: str

    #     with pytest.raises(
    #         ValueError, match="(Received `instructions` but this model)"
    #     ):
    #         Test("Hello!", instructions="Translate to French")
    #     with pytest.raises(ValueError, match="(Received `model` but this model)"):
    #         Test("Hello!", model=None)

    def test_instructions(self):
        @ai_model
        class Text(BaseModel):
            text: str

        t1 = Text("Hello")
        assert t1.text == "Hello"

        # this model is identical except it has an instruction
        @ai_model(instructions="first translate the text to French")
        class Text(BaseModel):
            text: str

        t2 = Text("Hello")
        assert t2.text == "Bonjour"

    def test_follow_instance_instructions(self):
        @ai_model
        class Test(BaseModel):
            text: str

        t1 = Test("Hello")
        assert t1.text == "Hello"

        # this model is identical except it has an instruction
        @ai_model
        class Test(BaseModel):
            text: str

        t2 = Test("Hello", instructions_="first translate the text to French")
        assert t2.text == "Bonjour"

    def test_follow_global_and_instance_instructions(self):
        @ai_model(instructions="Always set color_1 to 'red'")
        class Test(BaseModel):
            color_1: str
            color_2: str

        t1 = Test("Hello", instructions_="Always set color_2 to 'blue'")
        assert t1 == Test(color_1="red", color_2="blue")

    def test_follow_docstring_and_global_and_instance_instructions(self):
        @ai_model(instructions="Always set color_1 to 'red'")
        class Test(BaseModel):
            """Always set color_3 to 'orange'"""

            color_1: str
            color_2: str
            color_3: str

        t1 = Test("Hello", instructions_="Always set color_2 to 'blue'")
        assert t1 == Test(color_1="red", color_2="blue", color_3="orange")

    def test_follow_multiple_instructions(self):
        # ensure that instructions don't bleed to other invocations
        @ai_model
        class Translation(BaseModel):
            """Translates from one language to another language"""

            original_text: str
            translated_text: str

        t1 = Translation("Hello, world!", instructions_="Translate to French")
        t2 = Translation("Hello, world!", instructions_="Translate to German")

        assert t1 == Translation(
            original_text="Hello, world!", translated_text="Bonjour, monde!"
        )
        assert t2 == Translation(
            original_text="Hello, world!", translated_text="Hallo, Welt!"
        )


@pytest_mark_class("llm")
class TestAIModelMapping:
    # def test_arithmetic(self):
    #     @ai_model
    #     class Arithmetic(BaseModel):
    #         sum: float

    #     x = Arithmetic.map(["One plus six", "Two plus 100 minus one"])
    #     assert len(x) == 2
    #     assert x[0].sum == 7
    #     assert x[1].sum == 101

    def test_location(self):
        @ai_model
        class City(BaseModel):
            name: str = Field(description="The correct city name, e.g. Omaha")  # noqa

        results = City.map(
            [
                "the windy city",
                "chicago IL",
                "Chicago",
                "Chcago",
                "chicago, Illinois, USA",
                "chi-town",
            ]
        )
        assert len(results) == 6
        for result in results:
            assert result.name == "Chicago"

    def test_instructions(self):
        @ai_model
        class Translate(BaseModel):
            text: str

        result = Translate.map(["Hello", "Goodbye"], instructions="Translate to French")
        assert len(result) == 2
        assert result[0].text == "Bonjour"
        assert result[1].text == "Au revoir"



from datetime import datetime

from marvin.components.ai_function import ai_fn


@ai_fn
def summarize_text(text: str, specifications: str = "concise, comprehensive") -> str:
    """generates a summary of `text` according to the `specifications`"""


@ai_fn
def make_datetime(description: str, tz: str = "UTC") -> datetime:
    """generates a datetime from a description"""



from typing import Optional, Any, TypeVar
from pydantic import BaseModel
from .abstract import AbstractChatCompletion
from marvin.settings import settings

T = TypeVar(
    "T",
    bound=BaseModel,
)

PROVIDER_SHORTCUTS = {
    "gpt-3.5-turbo": "openai",
    "gpt-4": "openai",
    "claude-1": "anthropic",
    "claude-2": "anthropic",
}


def parse_model_shortcut(provider: Optional[str]) -> tuple[str, str]:
    """
    Parse a model string into a provider and a model name.
    - If the provider is None, use the default provider and model.
    - If the provider is a shortcut, use the shortcut to get the provider and model.
    """
    if provider is None:
        try:
            provider, model = settings.llm_model.split("/")
        except Exception:
            provider, model = (
                PROVIDER_SHORTCUTS[str(settings.llm_model)],
                settings.llm_model,
            )

    elif provider in PROVIDER_SHORTCUTS:
        provider, model = PROVIDER_SHORTCUTS[provider], provider
    else:
        provider, model = provider.split("/")
    return provider, model


def ChatCompletion(
    model: Optional[str] = None,
    **kwargs: Any,
) -> AbstractChatCompletion[T]:  # type: ignore
    provider, model = parse_model_shortcut(model)
    if provider == "openai" or provider == "azure_openai":
        from .providers.openai import OpenAIChatCompletion

        return OpenAIChatCompletion(provider=provider, model=model, **kwargs)
    if provider == "anthropic":
        from .providers.anthropic import AnthropicChatCompletion

        return AnthropicChatCompletion(model=model, **kwargs)
    else:
        raise ValueError(f"Unknown provider: {provider}")



from .ai_function import ai_fn
from .ai_function import AIFunction
from .ai_application import AIApplication
from .ai_model import ai_model
from .ai_model_factory import AIModelFactory
from .ai_model import AIModel
from .ai_classifier import ai_classifier



import inspect
import json
from ast import literal_eval
from typing import Any, Callable, List, Optional, Union

from pydantic import Field

import marvin
from marvin.engine.language_models import OpenAIFunction
from marvin.utilities.messages import Message

from .base import Executor


class OpenAIFunctionsExecutor(Executor):
    """
    An executor that understands how to pass functions to the LLM, interpret
    responses that request function calls, and iteratively continue to process
    functions until the LLM responds directly to the user. This uses the OpenAI
    Functions API, so provider LLMs must be compatible.
    """

    functions: Optional[List[OpenAIFunction]] = Field(default=None)
    function_call: Union[str, dict[str, str]] = Field(default=None)
    max_iterations: Optional[int] = Field(
        default_factory=lambda: marvin.settings.ai_application_max_iterations
    )
    stream_handler: Optional[Callable[[Message], None]] = Field(default=None)

    def __init__(
        self,
        functions: Optional[List[Union[OpenAIFunction, Callable[..., Any]]]] = None,
        **kwargs: Any,
    ):
        if functions is not None:
            functions = [
                (
                    OpenAIFunction.from_function(i)
                    if not isinstance(i, OpenAIFunction)
                    else i
                )
                for i in functions
            ]
        super().__init__(
            functions=functions,
            **kwargs,
        )

    # @validator("functions", pre=True)
    # def validate_functions(cls, v):
    #     if v is None:
    #         return None
    #     v = [
    #         OpenAIFunction.from_function(i) if not isinstance(i, OpenAIFunction)else i
    #         for i in v
    #     ]
    #     return v

    async def run_engine(self, messages: list[Message]) -> Message:
        """
        Implements one step of the LLM loop
        """

        kwargs = {}

        if self.functions:
            kwargs["functions"] = self.functions
            kwargs["function_call"] = self.function_call

        llm_response = await self.model.run(
            messages=messages,
            stream_handler=self.stream_handler,
            **kwargs,
        )
        return llm_response

    async def stop_condition(
        self, messages: List[Message], responses: List[Message]
    ) -> bool:
        # if the number of responses exceeds max iterations, stop

        if self.max_iterations is not None and len(responses) >= self.max_iterations:
            return True

        # if function calls are set to auto and the most recent call was a
        # function, continue
        if self.function_call == "auto":
            if responses and responses[-1].role == "function_response":
                return False

        # if a specific function call was requested but errored, continue
        elif self.function_call != "none":
            if responses and responses[-1].data.get("is_error"):
                return False

        # otherwise stop
        return True

    async def process_response(self, response: Message) -> Message:
        if response.role == "function_request":
            return await self.process_function_call(response)
        else:
            return response

    async def process_function_call(self, response: Message) -> Message:
        response_data = {}

        function_call = response.data["function_call"]
        fn_name = function_call.get("name")
        fn_args = function_call.get("arguments")
        response_data["name"] = fn_name
        try:
            try:
                fn_args = json.loads(function_call.get("arguments", "{}"))
            except json.JSONDecodeError:
                fn_args = literal_eval(function_call.get("arguments", "{}"))
            response_data["arguments"] = fn_args

            # retrieve the named function
            openai_fn = next((f for f in self.functions if f.name == fn_name), None)
            if openai_fn is None:
                raise ValueError(f'Function "{function_call["name"]}" not found.')

            if not isinstance(fn_args, dict):
                raise ValueError(
                    "Expected a dictionary of arguments, got a"
                    f" {type(fn_args).__name__}."
                )

            # call the function
            if openai_fn.fn is not None:
                self.logger.debug(
                    f"Running function '{openai_fn.name}' with payload {fn_args}"
                )
                fn_result = openai_fn.fn(**fn_args)
                if inspect.isawaitable(fn_result):
                    fn_result = await fn_result

            # if the function is undefined, return the arguments as its output
            else:
                fn_result = fn_args
            self.logger.debug(f"Result of function '{openai_fn.name}': {fn_result}")
            response_data["is_error"] = False

        except Exception as exc:
            fn_result = (
                f"The function '{fn_name}' encountered an error:"
                f" {str(exc)}\n\nThe payload you provided was: {fn_args}\n\nYou"
                " can try to fix the error and call the function again."
            )
            self.logger.debug_kv("Error", fn_result, key_style="red")
            response_data["is_error"] = True

        response_data["result"] = fn_result

        return Message(
            role="function_response",
            name=fn_name,
            content=str(fn_result),
            data=response_data,
            llm_response=response.llm_response,
        )



from pydantic.version import VERSION as PYDANTIC_VERSION

PYDANTIC_V2 = PYDANTIC_VERSION.startswith("2.")

if not PYDANTIC_V2:
    from pydantic import (
        BaseModel,
        BaseSettings,
        Extra,
        Field,
        PrivateAttr,
        PyObject,
        SecretStr,
        validate_arguments,
    )
    from pydantic.main import ModelMetaclass

    ModelMetaclass = ModelMetaclass
    BaseModel = BaseModel
    BaseSettings = BaseSettings
    Field = Field
    SecretStr = SecretStr
    Extra = Extra
    ImportString = PyObject
    PrivateAttr = PrivateAttr
    validate_arguments = validate_arguments



import asyncio
import inspect
from enum import Enum, EnumMeta  # noqa
from functools import partial
from typing import Any, Callable, Literal, Optional, TypeVar

from pydantic import BaseModel, Field
from typing_extensions import ParamSpec, Self

from marvin.core.ChatCompletion import ChatCompletion
from marvin.core.ChatCompletion.abstract import AbstractChatCompletion
from marvin.prompts import Prompt, prompt_fn
from marvin.utilities.async_utils import run_sync

T = TypeVar("T", bound=BaseModel)

A = TypeVar("A", bound=Any)

P = ParamSpec("P")


def ai_classifier_prompt(
    enum: Enum,
    ctx: Optional[dict[str, Any]] = None,
    **kwargs: Any,
) -> Callable[P, Prompt[P]]:
    @prompt_fn(
        ctx={"ctx": ctx or {}, "enum": enum, "inspect": inspect},
        response_model=int,  # type: ignore
        response_model_name="Index",
        response_model_description="The index of the most likely class.",
        response_model_field_name="index",
        serialize_on_call=False,
        **kwargs,
    )

    #     You are an expert classifier that always chooses correctly.
    #     {% if enum_class_docstring %}
    #     Your classification task is: {{ enum_class_docstring }}
    #     {% endif %}
    #     {% if instructions %}
    #     Your instructions are: {{ instructions }}
    #     {% endif %}
    #     The user will provide context through text, you will use your expertise
    #     to choose the best option below based on it:
    #     {% for option in options %}
    #         {{ loop.index }}. {{ value_getter(option) }}
    #     {% endfor %}
    #     {% if context_fn %}
    #     You have been provided the following context to perform your task:\n
    #     {%for (arg, value) in context_fn(value).items()%}
    #         - {{ arg }}: {{ value }}\n
    #     {% endfor %}
    #     {% endif %}\
    def prompt_wrapper(text: str) -> None:  # type: ignore # noqa
        """
        System: You are an expert classifier that always chooses correctly
        {{ '(note, however: ' + ctx.get('instructions') + ')' if ctx.get('instructions') }}

        {{ 'Also note that: ' + enum.__doc__ if enum.__doc__ }}

        The user will provide text to classify, you will use your expertise
        to choose the best option below based on it:
        {% for option in enum %}
            {{ loop.index }}. {{option.name}} ({{option.value}})
        {% endfor %}
        {% set context = ctx.get('context_fn')(text).items() if ctx.get('context_fn') %}
        {% if context %}
        You have been provided the following context to perform your task:
        {%for (arg, value) in context%}
            - {{ arg }}: {{ value }}\n
        {% endfor %}
        {% endif %}
        User: the text to classify: {{text}}
        """  # noqa

    return prompt_wrapper  # type: ignore


class AIEnumMetaData(BaseModel):
    model: Any = Field(default_factory=ChatCompletion)
    ctx: Optional[dict[str, Any]] = None
    instructions: Optional[str] = None
    mode: Optional[Literal["function", "logit_bias"]] = "logit_bias"


class AIEnumMeta(EnumMeta):
    """

    A metaclass for the AIEnum class.

    Enables overloading of the __call__ method to permit extra keyword arguments.

    """

    __metadata__ = AIEnumMetaData()

    def __call__(
        cls: Self,
        value: Any,
        names: Optional[Any] = None,
        *args: Any,
        module: Optional[str] = None,
        qualname: Optional[str] = None,
        type: Optional[type] = None,
        start: int = 1,
        boundary: Optional[Any] = None,
        model: Optional[str] = None,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        mode: Optional[Literal["function", "logit_bias"]] = None,
        **model_kwargs: Any,
    ) -> type[Enum]:
        cls.__metadata__ = AIEnumMetaData(
            model=ChatCompletion(model=model, **model_kwargs),
            ctx=ctx,
            instructions=instructions,
            mode=mode,
        )
        return super().__call__(
            value,
            names,  # type: ignore
            *args,
            module=module,
            qualname=qualname,
            type=type,
            start=start,
        )


class AIEnum(Enum, metaclass=AIEnumMeta):
    """
    AIEnum is a class that extends Python's built-in Enum class.
    It uses the AIEnumMeta metaclass, which allows additional parameters to be passed
    when creating an enum. These parameters are used to customize the behavior
    of the AI classifier.
    """

    @classmethod
    def _missing_(cls: type[Self], value: object) -> Self:
        response: int = cls.call(value)
        return list(cls)[response - 1]

    @classmethod
    def get_prompt(
        cls,
        *,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        **kwargs: Any,
    ) -> Callable[..., Prompt[P]]:
        ctx = ctx or cls.__metadata__.ctx or {}
        instructions = instructions or cls.__metadata__.instructions
        ctx["instructions"] = instructions or ctx.get("instructions", None)
        return ai_classifier_prompt(cls, ctx=ctx, **kwargs)  # type: ignore # noqa

    @classmethod
    def as_prompt(
        cls,
        value: Any,
        *,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        mode: Optional[Literal["function", "logit_bias"]] = None,
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> dict[str, Any]:
        ctx = ctx or cls.__metadata__.ctx or {}
        instructions = instructions or cls.__metadata__.instructions
        ctx["instructions"] = instructions or ctx.get("instructions", None)
        mode = mode or cls.__metadata__.mode
        response = cls.get_prompt(instructions=instructions, ctx=ctx)(value).serialize(
            model=cls.__metadata__.model,
        )
        if mode == "logit_bias":
            import tiktoken

            response.pop("functions", None)
            response.pop("function_call", None)
            response.pop("response_model", None)
            encoder = tiktoken.encoding_for_model("gpt-3.5-turbo")
            response["logit_bias"] = {
                encoder.encode(str(j))[0]: 100 for j in range(1, len(cls) + 1)
            }
            response["max_tokens"] = 1
        return response

    @classmethod
    def as_dict(
        cls,
        value: Any,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        mode: Optional[Literal["function", "logit_bias"]] = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        ctx = ctx or cls.__metadata__.ctx or {}
        instructions = instructions or cls.__metadata__.instructions
        ctx["instructions"] = instructions or ctx.get("instructions", None)
        mode = mode or cls.__metadata__.mode

        response = cls.get_prompt(ctx=ctx, instructions=instructions)(value).to_dict()
        if mode == "logit_bias":
            import tiktoken

            response.pop("functions", None)
            response.pop("function_call", None)
            response.pop("response_model", None)
            encoder = tiktoken.encoding_for_model("gpt-3.5-turbo")
            response["logit_bias"] = {
                encoder.encode(str(j))[0]: 100 for j in range(1, len(cls) + 1)
            }
            response["max_tokens"] = 1

        return response

    @classmethod
    def as_chat_completion(
        cls,
        value: Any,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        mode: Optional[Literal["function", "logit_bias"]] = None,
    ) -> AbstractChatCompletion[T]:  # type: ignore # noqa
        ctx = ctx or cls.__metadata__.ctx or {}
        instructions = instructions or cls.__metadata__.instructions
        mode = mode or cls.__metadata__.mode
        ctx["instructions"] = instructions or ctx.get("instructions", None)
        return cls.__metadata__.model(
            **cls.as_dict(value, ctx=ctx, instructions=instructions, mode=mode)
        )

    @classmethod
    def call(
        cls,
        value: Any,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        mode: Optional[Literal["function", "logit_bias"]] = None,
    ) -> Any:
        ctx = ctx or cls.__metadata__.ctx or {}
        instructions = instructions or cls.__metadata__.instructions
        ctx["instructions"] = instructions or ctx.get("instructions", None)
        mode = mode or cls.__metadata__.mode
        chat_completion = cls.as_chat_completion(  # type: ignore # noqa
            value, ctx=ctx, instructions=instructions, mode=mode
        )
        if cls.__metadata__.mode == "logit_bias":
            return int(chat_completion.create().response.choices[0].message.content)  # type: ignore # noqa
        return getattr(chat_completion.create().to_model(), "index")  # type: ignore

    @classmethod
    async def acall(
        cls,
        value: Any,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        mode: Optional[Literal["function", "logit_bias"]] = None,
    ) -> Any:
        ctx = ctx or cls.__metadata__.ctx or {}
        instructions = instructions or cls.__metadata__.instructions
        ctx["instructions"] = instructions or ctx.get("instructions", None)
        mode = mode or cls.__metadata__.mode
        chat_completion = cls.as_chat_completion(  # type: ignore # noqa
            value, ctx=ctx, instructions=instructions, mode=mode
        )
        if cls.__metadata__.mode == "logit_bias":
            return int((await chat_completion.acreate()).response.choices[0].message.content)  # type: ignore # noqa
        return getattr((await chat_completion.acreate()).to_model(), "index")  # type: ignore # noqa

    @classmethod
    def map(cls, items: list[str], **kwargs: Any) -> list[Any]:
        """
        Map the classifier over a list of items.
        """
        coros = [cls.acall(item, **kwargs) for item in items]

        # gather returns a future, but run_sync requires a coroutine
        async def gather_coros() -> list[Any]:
            return await asyncio.gather(*coros)

        results = run_sync(gather_coros())
        return [list(cls)[result - 1] for result in results]

    @classmethod
    def as_decorator(
        cls: type[Self],
        enum: Optional[Enum] = None,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        mode: Optional[Literal["function", "logit_bias"]] = "logit_bias",
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> Self:
        if not enum:
            return partial(
                cls.as_decorator,
                ctx=ctx,
                instructions=instructions,
                mode=mode,
                model=model,
                **model_kwargs,
            )  # type: ignore
        response = cls(
            enum.__name__,  # type: ignore
            {member.name: member.value for member in enum},  # type: ignore
        )
        setattr(
            response,
            "__metadata__",
            AIEnumMetaData(
                model=ChatCompletion(model=model, **model_kwargs),
                ctx=ctx,
                instructions=instructions,
                mode=mode,
            ),
        )

        response.__doc__ = enum.__doc__  # type: ignore
        return response


ai_classifier = AIEnum.as_decorator



from pathlib import Path

import marvin
import marvin.tools
import marvin.tools.filesystem
import marvin.tools.python
import marvin.tools.shell
from marvin import AIApplication

marvin.settings.log_level = "DEBUG"
marvin.settings.llm_model = "gpt-4"


ROOT_DIR = Path(marvin.__file__).parents[2]
DOCS_DIR = ROOT_DIR / "docs"
MKDOCS_FILE = ROOT_DIR / "mkdocs.yml"

docs_app = AIApplication(
    name="DocsWriter",
    description=f"""
        You are an expert technical writer, responsible for maintaining high
        quality documentation for the Marvin library. 
        
        Marvin is an AI engineering framework written in Python. Its
        documentation is based on Material for MKDocs and has a mix of
        conceptual guides, tutorials, and reference documentation. You can
        update and create new documentation as necessary, and be sure to update
        the `mkdocs.yml` file's navigation section appropriately. 
        
        When crafting Marvin's documentation, your aim should be for a tone that
        is professional yet conversational. Our writing style prioritizes
        clarity and understandability, utilizing plain English and circumventing
        unnecessary jargon. We believe in demystifying complex concepts through
        the use of metaphors and analogies, likening Marvin to a "Swiss Army
        Knife" to denote its multitude of functions. Your content should be
        structured into digestible chunks, with each segment focusing on a
        singular idea or topic. Effective organization of your sections through
        headers, listing details with bullet points, and conveying technical
        specifications with code blocks are highly encouraged. Hyperlinks are
        integral in guiding readers to pertinent content both within and outside
        the document. Aim to establish a sense of community by directly
        addressing our readers with words like "we" and "you." Marvin's
        documentation is not merely a technical manual; it mirrors our
        inclusive, innovative, and audacious personality. While we do value
        humor and wit, referencing Marvin's cultural background in the
        Hitchhiker's Guide to the Galaxy, it should be used sparingly and
        fittingly to maintain its charm and effectiveness. Remember, our goal
        with our writing is not just to inform, but to engage, inspire, and
        occasionally elicit a smile from our readers.
        
        The library source is available at {ROOT_DIR}. The documentation is
        contained in {DOCS_DIR}. The `mkdocs.yml` file is located at
        {MKDOCS_FILE}.
        """,
    tools=[
        marvin.tools.filesystem.ListFiles(root_dir=ROOT_DIR),
        marvin.tools.filesystem.ReadFile(root_dir=ROOT_DIR),
        # marvin.tools.filesystem.ReadFiles(root_dir=ROOT_DIR),
        # marvin.tools.filesystem.WriteFile(
        #     root_dir=ROOT_DIR, require_confirmation=False
        # ),
        marvin.tools.filesystem.WriteFiles(
            root_dir=ROOT_DIR, require_confirmation=False
        ),
        marvin.tools.python.Python(require_confirmation=False),
        marvin.tools.shell.Shell(
            require_confirmation=False, working_directory=ROOT_DIR
        ),
    ],
)



from marvin import AIApplication


class Agent(AIApplication):
    description: str = "A helpful AI assistant"

    def __init__(self, **kwargs):
        super().__init__(
            state_enabled=False,
            plan_enabled=False,
            **kwargs,
        )


__all__ = ["Agent"]



from pydantic import BaseModel


class DefaultSettings(BaseModel):
    default_model_path: str = "marvin.language_models.default"
    default_model_name: str = "gpt-4"
    default_model_api_key_name: str = "OPENAI_API_KEY"


default_settings = DefaultSettings().dict()



import sys
from io import StringIO

from marvin.tools import Tool


def run_python(code: str, globals: dict = None, locals: dict = None):
    old_stdout = sys.stdout
    new_stdout = StringIO()
    sys.stdout = new_stdout
    try:
        exec(code, globals or {}, locals or {})
        result = new_stdout.getvalue()
    except Exception as e:
        result = repr(e)
    finally:
        sys.stdout = old_stdout
    return result


class Python(Tool):
    description: str = """
    Runs arbitrary Python code.

    {%if require_confirmation %} You MUST ask the user to confirm execution by
    showing them the code. {% endif %}
    """
    require_confirmation: bool = True

    def run(self, code: str) -> str:
        return run_python(code)



from pydantic import BaseModel, Field, validator, Extra, BaseSettings, root_validator
from pydantic.main import ModelMetaclass

from typing import Any, Callable, List, Optional, Type, Union, Literal
from marvin import settings
from marvin.types import Function
from operator import itemgetter
from marvin.utilities.module_loading import import_string
import warnings
import copy
from marvin.types.request import Request as BaseRequest
from marvin.engine import ChatCompletionBase


class Request(BaseRequest):
    """
    This is a class for creating Request objects to interact with the GPT-3 API.
    The class contains several configurations and validation functions to ensure
    the correct data is sent to the API.

    """

    model: str = "gpt-3.5-turbo"  # the model used by the GPT-3 API
    temperature: float = 0.8  # the temperature parameter used by the GPT-3 API
    api_key: str = Field(default_factory=settings.openai.api_key.get_secret_value)

    class Config:
        exclude = {"response_model"}
        exclude_none = True
        extra = Extra.allow

    def dict(self, *args, serialize_functions=True, exclude=None, **kwargs):
        """
        This method returns a dictionary representation of the Request.
        If the functions attribute is present and serialize_functions is True,
        the functions' schemas are also included.
        """

        # This identity function is here for no reason except to show
        # readers that custom adapters need only override the dict method.
        return super().dict(
            *args, serialize_functions=serialize_functions, exclude=exclude, **kwargs
        )


class Response(BaseModel):
    """
    This class is used to handle the response from the API.
    It includes several utility functions and properties to extract useful information
    from the raw response.
    """

    raw: Any  # the raw response from the API
    request: Any  # the request that generated the response

    def __init__(self, response, *args, request, **kwargs):
        super().__init__(raw=response, request=request)

    def __iter__(self):
        return self.raw.__iter__()

    def __next__(self):
        return self.raw.__next__()

    def __getattr__(self, name):
        """
        This method attempts to get the attribute from the raw response.
        If it doesn't exist, it falls back to the standard attribute access.
        """
        try:
            return self.raw.__getattr__(name)
        except AttributeError:
            return self.__getattribute__(name)

    @property
    def message(self):
        """
        This property extracts the message from the raw response.
        If there is only one choice, it returns the message from that choice.
        Otherwise, it returns a list of messages from all choices.
        """
        if len(self.raw.choices) == 1:
            return next(iter(self.raw.choices)).message
        return [x.message for x in self.raw.choices]

    @property
    def function_call(self):
        """
        This property extracts the function call from the message.
        If the message is a list, it returns a list of function calls from all messages.
        Otherwise, it returns the function call from the message.
        """
        if isinstance(self.message, list):
            return [x.function_call for x in self.message]
        return self.message.function_call

    @property
    def callables(self):
        """
        This property returns a list of all callable functions from the request.
        """
        return [x for x in self.request.functions if isinstance(x, Callable)]

    @property
    def callable_registry(self):
        """
        This property returns a dictionary mapping function names to functions for all
        callable functions from the request.
        """
        return {fn.__name__: fn for fn in self.callables}

    def call_function(self, as_message=True):
        """
        This method evaluates the function call in the response and returns the result.
        If as_message is True, it returns the result as a function message.
        Otherwise, it returns the result directly.
        """
        name, raw_arguments = itemgetter("name", "arguments")(self.function_call)
        function = self.callable_registry.get(name)
        arguments = function.model.parse_raw(raw_arguments)
        value = function(**arguments.dict(exclude_none=True))
        if as_message:
            return {"role": "function", "name": name, "content": value}
        else:
            return value

    def to_model(self):
        """
        This method parses the function call arguments into the response model and
        returns the result.
        """
        return self.request.response_model.parse_raw(self.function_call.arguments)

    def __repr__(self, *args, **kwargs):
        """
        This method returns a string representation of the raw response.
        """
        return self.raw.__repr__(*args, **kwargs)


class OpenAIChatCompletion(ChatCompletionBase):
    """
    This class is used to create and handle chat completions from the API.
    It provides several utility functions to create the request, send it to the API,
    and handle the response.
    """

    _module: str = "openai.ChatCompletion"  # the module used to interact with the API
    _request: str = "marvin.openai.ChatCompletion.Request"
    _response: str = "marvin.openai.ChatCompletion.Response"
    defaults: Optional[dict] = Field(None, repr=False)  # default configuration values


ChatCompletion = OpenAIChatCompletion()

# This is a legacy class that is used to create a ChatCompletion object.
# It is deprecated and will be removed in a future release.
ChatCompletionConfig = Request



import subprocess
from pathlib import Path

from marvin.tools import Tool


class Shell(Tool):
    description: str = """
    Runs arbitrary shell code.
    
    {% if working_directory %} The working directory will be {{
    working_directory }}. {% endif %}.

    {%if require_confirmation %} You MUST ask the user to confirm execution by
    showing them the code. {% endif %}
    """
    require_confirmation: bool = True
    working_directory: Path = None

    def run(self, cmd: str, working_directory: str = None) -> str:
        if working_directory and self.working_directory:
            raise ValueError(
                f"The working directoy is {self.working_directory}; do not provide"
                " another one."
            )

        wd = working_directory or (
            str(self.working_directory) if self.working_directory else None
        )

        try:
            result = subprocess.run(
                cmd,
                shell=True,
                cwd=wd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            if result.returncode != 0:
                return (
                    f"Command failed with code {result.returncode} and error:"
                    f" {result.stderr.decode() or '<No error>'}"
                )
            else:
                return (
                    "Command succeeded with output:"
                    f" { result.stdout.decode() or '<No output>' }"
                )

        except Exception as e:
            return f"Execution failed: {str(e)}"



from chatbot import handle_message
from marvin import AIApplication
from marvin.deployment import Deployment

deployment = Deployment(
    component=AIApplication(tools=[handle_message]),
    app_kwargs={
        "title": "Marvin Slackbot",
        "description": "A Slackbot powered by Marvin",
    },
    uvicorn_kwargs={
        "port": 4200,
    },
)

if __name__ == "__main__":
    deployment.serve()



from .typer import AsyncTyper

from .admin import app as admin
from .chat import chat

app = AsyncTyper()

app.add_typer(admin, name="admin")

app.acommand()(chat)


@app.command()
def version():
    import platform
    import sys
    from marvin import __version__

    print(f"Version:\t\t{__version__}")

    print(f"Python version:\t\t{sys.version.split()[0]}")

    print(f"OS/Arch:\t\t{platform.system().lower()}/{platform.machine().lower()}")


if __name__ == "__main__":
    app()



from fastapi import FastAPI, Request
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

app = FastAPI()

# Mount the static files directory
app.mount("/static", StaticFiles(directory="static"), name="static")

# Mount the templates directory
templates = Jinja2Templates(directory="templates")


@app.get("/")
async def root(request: Request):
    # Return the index.html file
    return templates.TemplateResponse("index.html", {"request": request})



from typing import Callable, List, Union

from marvin import AIApplication
from marvin.prompts import Prompt
from marvin.tools.base import Tool

DEFAULT_NAME = "Marvin"
DEFAULT_PERSONALITY = "A friendly AI assistant"
DEFAULT_INSTRUCTIONS = "Engage the user in conversation."


class Chatbot(AIApplication):
    name: str = DEFAULT_NAME
    personality: str = DEFAULT_PERSONALITY
    instructions: str = DEFAULT_INSTRUCTIONS
    tools: List[Union[Tool, Callable]] = ([],)

    def __init__(
        self,
        name: str = DEFAULT_NAME,
        personality: str = DEFAULT_PERSONALITY,
        instructions: str = DEFAULT_INSTRUCTIONS,
        state=None,
        tools: list[Union[Tool, Callable]] = [],
        additional_prompts: list[Prompt] = None,
        **kwargs,
    ):
        description = f"""
            You are a chatbot - your name is {name}.
            
            You must respond to the user in accordance with
            your personality and instructions.
            
            Your personality is: {personality}.
            
            Your instructions are: {instructions}.
            """
        super().__init__(
            name=name,
            description=description,
            tools=tools,
            state=state or {},
            state_enabled=False if state is None else True,
            plan_enabled=False,
            additional_prompts=additional_prompts or [],
            **kwargs,
        )


__all__ = ["Chatbot"]



import functools
import inspect
import re
from typing import Callable, TypeVar, Optional, Any, Type

from fastapi.routing import APIRouter

from marvin.utilities.types import function_to_model


T = TypeVar("T")
A = TypeVar("A")


class Function:
    def __init__(
        self, *, fn: Callable[[A], T] = None, name: str = None, description: str = None
    ) -> None:
        self.fn = fn
        self.name = name or self.fn.__name__
        self.description = description or self.fn.__doc__

        super().__init__()

    @property
    def model(self):
        return function_to_model(self.fn, name=self.name, description=self.description)

    @property
    def signature(self):
        return inspect.signature(self.fn)

    @property
    def source_code(self):
        source_code = inspect.cleandoc(inspect.getsource(self.fn))
        if match := re.search(re.compile(r"(\bdef\b.*)", re.DOTALL), source_code):
            source_code = match.group(0)
        return source_code

    @property
    def return_annotation(self):
        return_annotation = self.signature.return_annotation
        if return_annotation is inspect._empty:
            return return_annotation, False
        return return_annotation

    def arguments(self, *args, **kwargs):
        bound_args = self.signature.bind(*args, **kwargs)
        bound_args.apply_defaults()
        return bound_args.arguments

    def schema(self, *args, name: str = None, description: str = None, **kwargs):
        schema = self.model.schema(*args, **kwargs)
        return {
            "name": name or schema.pop("title"),
            "description": description or self.fn.__doc__,
            "parameters": schema,
        }


def FunctionDecoratorFactory(
    name: str = "marvin",
    func_class: Type[T] = None,
    in_place=True,
) -> Callable[[A], T]:
    def decorator(fn: Callable[[A], T] = None) -> Callable[[A], T]:
        if fn is None:
            return functools.partial(decorator)
        elif in_place:
            fn = func_class(fn=fn)
        else:
            instance = func_class(fn=fn)
            setattr(fn, name, instance)
            for method in dir(instance):
                is_method_private = method.startswith("__")
                if not is_method_private:
                    setattr(fn, method, getattr(instance, method))
        return fn

    return decorator


marvin_fn = FunctionDecoratorFactory(name="openai", func_class=Function)


class FunctionRegistry(APIRouter):
    def __init__(self, function_decorator=marvin_fn, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.function_decorator = function_decorator

    @property
    def endpoints(self):
        # Returns literal functions.
        return [route.endpoint for route in self.routes]

    @property
    def schema(self):
        # Returns JSON Schema of functions.
        return [self.function_decorator(fn=fn).schema() for fn in self.endpoints]

    @property
    def functions(self):
        # Returns function classes.
        return [self.function_decorator(fn=fn) for fn in self.endpoints]

    def include(self, registry: "FunctionRegistry", *args, **kwargs):
        super().include_router(registry, *args, **kwargs)
        # Add some 50-IQ idempotency.
        self.routes = list({x.name: x for x in self.routes}.values())

    def register(self, fn: Optional[Callable] = None, **kwargs: Any) -> Callable:
        def decorator(fn: Callable, *args) -> Callable:
            fn_class = self.function_decorator(fn=fn, **kwargs)
            self.add_api_route(
                **{
                    **{
                        "name": fn_class.name,
                        "path": f"/{fn_class.name}",
                        "endpoint": fn,
                        "description": fn_class.description,
                        "methods": ["POST"],
                    },
                    **kwargs,
                }
            )
            return fn

        if fn:
            # if the decorator was called with parentheses
            return decorator(fn)
        else:
            # else, return the decorator to be called later
            return decorator



from datetime import datetime
from typing import List, Optional

import httpx
from pydantic import BaseModel, Field, validator

import marvin
from marvin.tools import Tool
from marvin.utilities.strings import slice_tokens


class GitHubUser(BaseModel):
    """GitHub user."""

    login: str


class GitHubComment(BaseModel):
    """GitHub comment."""

    body: str = Field(default="")
    user: GitHubUser = Field(default_factory=GitHubUser)


class GitHubLabel(BaseModel):
    """GitHub label."""

    name: str = Field(default="")


class GitHubIssue(BaseModel):
    """GitHub issue."""

    created_at: datetime = Field(...)
    html_url: str = Field(...)
    number: int = Field(...)
    title: str = Field(default="")
    body: Optional[str] = Field(default="")
    labels: List[GitHubLabel] = Field(default_factory=GitHubLabel)
    user: GitHubUser = Field(default_factory=GitHubUser)

    @validator("body", always=True)
    def validate_body(cls, v):
        if not v:
            return ""
        return v


class SearchGitHubIssues(Tool):
    """Tool for searching GitHub issues."""

    description: str = "Use the GitHub API to search for issues in a given repository."

    async def run(self, query: str, repo: str = "prefecthq/prefect", n: int = 3) -> str:
        """
        Use the GitHub API to search for issues in a given repository. Do
        not alter the default value for `n` unless specifically requested by
        a user.

        For example, to search for open issues about AttributeErrors with the
        label "bug" in PrefectHQ/prefect:
            - repo: prefecthq/prefect
            - query: label:bug is:open AttributeError
        """
        headers = {"Accept": "application/vnd.github.v3+json"}

        if token := marvin.settings.github_token:
            headers["Authorization"] = f"Bearer {token.get_secret_value()}"

        async with httpx.AsyncClient() as client:
            response = await client.get(
                "https://api.github.com/search/issues",
                headers=headers,
                params={
                    "q": query if "repo:" in query else f"repo:{repo} {query}",
                    "order": "desc",
                    "per_page": n,
                },
            )
            response.raise_for_status()

        issues_data = response.json()["items"]

        # enforce 1000 token limit per body
        for issue in issues_data:
            if not issue["body"]:
                continue
            issue["body"] = slice_tokens(issue["body"], 1000)

        issues = [GitHubIssue(**issue) for issue in issues_data]

        summary = "\n\n".join(
            f"{issue.title} ({issue.html_url}):\n{issue.body}" for issue in issues
        )
        if not summary.strip():
            raise ValueError("No issues found.")
        return summary



from typing import List, Union

from pydantic import PrivateAttr

from marvin.engine.language_models import ChatLLM
from marvin.prompts.base import Prompt, render_prompts
from marvin.utilities.messages import Message
from marvin.utilities.types import LoggerMixin, MarvinBaseModel


class Executor(LoggerMixin, MarvinBaseModel):
    model: ChatLLM
    _should_stop: bool = PrivateAttr(False)

    async def start(
        self,
        prompts: list[Union[Prompt, Message]],
        prompt_render_kwargs: dict = None,
    ) -> list[Message]:
        """
        Start the LLM loop
        """
        # reset stop criteria
        self._should_stop = False

        responses = []
        while not self._should_stop:
            # render the prompts, including any responses from the previous step
            messages = render_prompts(
                prompts + responses,
                render_kwargs=prompt_render_kwargs,
                max_tokens=self.model.context_size,
            )
            response = await self.step(messages)
            responses.append(response)
            if await self.stop_condition(messages, responses):
                self._should_stop = True
        return responses

    async def step(self, messages: list[Message]) -> Message:
        """
        Implements one step of the LLM loop
        """
        messages = await self.process_messages(messages)
        llm_response = await self.run_engine(messages=messages)
        response = await self.process_response(llm_response)
        return response

    async def run_engine(self, messages: list[Message]) -> Message:
        """
        Implements one step of the LLM loop
        """
        llm_response = await self.model.run(messages=messages)
        return llm_response

    async def process_messages(self, messages: list[Message]) -> list[Message]:
        """Called prior to sending messages to the LLM"""
        return messages

    async def stop_condition(
        self, messages: List[Message], responses: List[Message]
    ) -> bool:
        return True

    async def process_response(self, response: Message) -> Message:
        """Called after receiving a response from the LLM"""
        return response



"""
The engine module is the interface to external LLM providers.
"""
from pydantic import BaseModel, Field
from typing import Optional
from marvin.utilities.module_loading import import_string
import copy


class ChatCompletionBase(BaseModel):
    """
    This class is used to create and handle chat completions from the API.
    It provides several utility functions to create the request, send it to the API,
    and handle the response.
    """

    _module: str = "openai.ChatCompletion"  # the module used to interact with the API
    _request: str = "marvin.openai.ChatCompletion.Request"
    _response: str = "marvin.openai.ChatCompletion.Response"
    _create: str = "create"  # the name of the create method in the API model
    _acreate: str = (  # the name of the asynchronous create method in the API model
        "acreate"
    )
    defaults: Optional[dict] = Field(None, repr=False)  # default configuration values

    def __init__(self, module_path: str = None, config_path: str = None, **kwargs):
        super().__init__(_module=module_path, _config=config_path, defaults=kwargs)

    @property
    def module(self):
        """
        This property imports and returns the API model.
        """
        return import_string(self._module)

    @property
    def model(self):
        """
        This property imports and returns the API model.
        """
        return self.module

    def request(self, *args, **kwargs):
        """
        This method imports and returns a configuration object.
        """
        return import_string(self._request)(*args, **(kwargs or self.defaults or {}))

    @property
    def response_class(self, *args, **kwargs):
        """
        This method imports and returns a configuration object.
        """
        return import_string(self._response)

    def prepare_request(self, *args, **kwargs):
        """
        This method prepares a request and returns it.
        """
        request = self.request() | self.request(**kwargs)
        payload = request.dict(exclude_none=True, exclude_unset=True)
        return request, payload

    def create(self=None, *args, **kwargs):
        """
        This method creates a request and sends it to the API.
        It returns a Response object with the raw response and the request.
        """
        request, request_dict = self.prepare_request(*args, **kwargs)
        create = getattr(self.model, self._create)
        response = self.response_class(create(**request_dict), request=request)
        if request.evaluate_function_call and response.function_call:
            return response.call_function(as_message=True)
        return response

    async def acreate(self, *args, **kwargs):
        """
        This method is an asynchronous version of the create method.
        It creates a request and sends it to the API asynchronously.
        It returns a Response object with the raw response and the request.
        """
        request, request_dict = self.prepare_request(*args, **kwargs)
        acreate = getattr(self.model, self._acreate)
        response = self.response_class(await acreate(**request_dict), request=request)
        if request.evaluate_function_call and response.function_call:
            return await response.acall_function(as_message=True)
        return response

    def __call__(self, *args, **kwargs):
        self = copy.deepcopy(self)
        request = self.request()
        passed = self.__class__(**kwargs).request()
        self.defaults = (request | passed).dict(serialize_functions=False)
        return self



import json
from ast import literal_eval
from typing import Any


def parse_raw(raw: str) -> dict[str, Any]:
    try:
        return literal_eval(raw)
    except Exception:
        pass
    try:
        return json.loads(raw)
    except Exception:
        pass
    return {}



import copy
import inspect
import re
from typing import Callable, Optional, Type

from pydantic import BaseModel

from marvin._compat import validate_arguments

extraneous_fields = [
    "args",
    "kwargs",
    "ALT_V_ARGS",
    "ALT_V_KWARGS",
    "V_POSITIONAL_ONLY_NAME",
    "V_DUPLICATE_KWARGS",
]


def get_openai_function_schema(schema, model):
    # Make a copy of the schema.
    _schema = copy.deepcopy(schema)

    # Prune the schema of 'titles'.
    _schema.pop("title", None)
    for key, value in schema["properties"].items():
        if key in extraneous_fields:
            _schema["properties"].pop(key, None)
        else:
            _schema["properties"][key].pop("title", None)

    # Clear the existing schema.
    schema.clear()

    # Reconstruct the schema.
    schema["name"] = getattr(model, "name", model.Config.fn.__name__)
    schema["description"] = getattr(model, "description", model.Config.fn.__doc__)
    schema["parameters"] = {
        k: v for (k, v) in _schema.items() if k not in extraneous_fields
    }


class FunctionConfig(BaseModel):
    fn: Callable
    name: str
    description: str = ""
    schema_extra: Optional[Callable] = get_openai_function_schema

    def __init__(self, fn, **kwargs):
        kwargs.setdefault("name", fn.__name__ or "")
        kwargs.setdefault("description", fn.__doc__ or "")
        super().__init__(fn=fn, **kwargs)

    def getsource(self):
        try:
            return re.search("def.*", inspect.getsource(self.fn), re.DOTALL).group()
        except Exception:
            return None

    def bind_arguments(self, *args, **kwargs):
        bound_arguments = inspect.signature(self.fn).bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        return bound_arguments.arguments

    def response_model(self, *args, **kwargs):
        def format_response(data: inspect.signature(self.fn).return_annotation):
            """Function to format the final response to the user"""
            return None

        format_response.__name__ = kwargs.get("name", format_response.__name__)
        format_response.__doc__ = kwargs.get("description", format_response.__doc__)
        response_model = Function(format_response).model
        response_model.__signature__ = inspect.signature(format_response)
        return response_model


class Function:
    """
    A wrapper class to add additional functionality to a function,
    such as a schema, response model, and more.
    """

    def __new__(cls, fn: Callable, parameters: dict = None, signature=None, **kwargs):
        config = FunctionConfig(fn, **kwargs)
        instance = super().__new__(cls)
        instance.instance = validate_arguments(fn, config=config.dict())
        instance.schema = parameters or instance.instance.model.schema
        instance.response_model = config.response_model
        instance.bind_arguments = config.bind_arguments
        instance.getsource = config.getsource
        instance.__name__ = config.name or fn.__name__
        instance.__doc__ = config.description or fn.__doc__
        instance.signature = signature or inspect.signature(fn)
        return instance

    def __call__(self, *args, **kwargs):
        return self.evaluate_raw(*args, **kwargs)

    def evaluate_raw(self, *args, **kwargs):
        return self.instance(*args, **kwargs)

    @classmethod
    def from_model(cls, model: Type[BaseModel], **kwargs):
        model.__signature__ = inspect.Signature(
            list(model.__signature__.parameters.values()), return_annotation=model
        )

        return cls(
            model, name="format_response", description="Format the response", **kwargs
        )

    @classmethod
    def from_return_annotation(
        cls, fn: Callable, *args, name: str = None, description: str = None
    ):
        def format_final_response(data: inspect.signature(fn).return_annotation):
            """Function to format the final response to the user"""
            return None

        format_final_response.__name__ = name or format_final_response.__name__
        format_final_response.__doc__ = description or format_final_response.__doc__
        return cls(format_final_response)

    def __repr__(self):
        parameters = []
        for _, param in self.signature.parameters.items():
            param_repr = str(param)
            if param.annotation is not param.empty:
                param_repr = param_repr.replace(
                    str(param.annotation),
                    (
                        param.annotation.__name__
                        if isinstance(param.annotation, type)
                        else str(param.annotation)
                    ),
                )
            parameters.append(param_repr)
        param_str = ", ".join(parameters)
        return f"marvin.functions.{self.__name__}({param_str})"



import json
import re
from typing import Any, Literal, Optional, Union

from jinja2 import Environment
from marvin.utilities.strings import jinja_env

from ...utils import parse_raw

FUNCTION_PROMPT = """
# Functions

You can call various functions to perform tasks.

Whenever you receive a message from the user, check to see if any of your
functions would help you respond. For example, you might use a function to look
up information, interact with a filesystem, call an API, or validate data. You
might write code, update state, or cause a side effect. After indicating that
you want to call a function, the user will execute the function and tell you its
result so that you can use the information in your final response. Therefore,
you must use your functions whenever they would be helpful.

The user may also provide a `function_call` instruction which could be:

- "auto": you may decide to call a function on your own (this is the
    default)
- "none": do not call any function
- {"name": "<function-name>"}: you MUST call the function with the given
    name

To call a function:

- Your response must include a JSON payload with the below format, including the
  {"mode": "function_call"} key.
- Do not put any other text in your response beside the JSON payload.
- Do not describe your plan to call the function to the user; they will not see
  it. 
- Do not include more than one payload in your response.
- Do not include function output or results in your response.
    
# Available Functions

Your have access to the following functions. Each has a name (which must be part
of your response), a description (which you should use to decide to call the
function), and a parameter spec (which is a JSON Schema description of the
arguments you should pass in your response)

{% for function in functions -%} 

## {{ function.name }} 

- Name: {{ function.name }} 
- Description: {{ function.description }} 
- Parameters: {{ function.parameters }}

{% endfor %}

# Calling a Function

To call a function, your response MUST include a JSON document with the
following structure: 

{
    "mode": "function_call", 
    
    "name": "<the name of the function, must be one of the above names>", 
    
    "arguments": "<a JSON string with the arguments to the function as valid
    JSON>"
}

The user will execute the function and respond with its result verbatim.

# function_call instruction

The user provided the following `function_call` instruction: {{ function_call }}

# final_response instruction

When you choose to call no functions, your final answer should be summative.
Do not acknowledge any internal functions you had access to or chose to call.

"""


def render_anthropic_functions_prompt(
    functions: list[dict[str, Any]],
    function_call: Union[dict[Literal["name"], str], Literal["auto"]],
    environment: Optional[Environment] = None,
) -> str:
    env = environment or jinja_env
    template = env.from_string(FUNCTION_PROMPT)
    return template.render(functions=functions, function_call=function_call or "auto")


def handle_anthropic_response(
    completion: str,
) -> tuple[Optional[str], Optional[dict[str, Any]]]:
    try:
        response: dict[str, Any] = parse_raw(
            re.findall(r"\{.*\}", completion, re.DOTALL)[0]
        )
        if response.pop("mode", None) == "function_call":
            return None, {
                "name": response.pop("name", None),
                "arguments": json.dumps(response.pop("arguments", None)),
            }
    except Exception:
        pass
    return completion, None



from typing import List

import openai

import marvin


async def create_openai_embeddings(texts: List[str]) -> List[List[float]]:
    """Create OpenAI embeddings for a list of texts."""

    try:
        import numpy  # noqa F401
    except ImportError:
        raise ImportError(
            "The numpy package is required to create OpenAI embeddings. Please install"
            " it with `pip install numpy` or `pip install 'marvin[slackbot]'`."
        )

    embeddings = await openai.Embedding.acreate(
        input=[text.replace("\n", " ") for text in texts],
        engine=marvin.settings.openai.embedding_engine,
    )

    return [
        r["embedding"] for r in sorted(embeddings["data"], key=lambda x: x["index"])
    ]



import asyncio
from collections.abc import Callable, Coroutine
from functools import wraps
from typing import Any, TypeVar

from typer import Typer
from typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")

# Comment from GitHub issue: https://github.com/tiangolo/typer/issues/88
# User: https://github.com/macintacos


class AsyncTyper(Typer):
    """Asyncronous Typer that derives from Typer.

    Use this when you have an asynchronous command you want to build,
    otherwise, just use Typer.
    """

    # Because we're being generic in this decorator, 'Any' is fine for the args.
    def acommand(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Callable[
        [Callable[P, Coroutine[Any, Any, R]]],
        Callable[P, Coroutine[Any, Any, R]],
    ]:
        """An async decorator for Typer commands that are asynchronous."""

        def decorator(
            async_func: Callable[P, Coroutine[Any, Any, R]],
        ) -> Callable[P, Coroutine[Any, Any, R]]:
            @wraps(async_func)
            def sync_func(*_args: P.args, **_kwargs: P.kwargs) -> R:
                return asyncio.run(async_func(*_args, **_kwargs))

            # Now use app.command as normal to register the synchronous function
            self.command(*args, **kwargs)(sync_func)

            # Return the async function unmodified, to preserved library functionality.
            return async_func

        return decorator
